from .. import *
from .base import *

class ComboBox(WxBaseWidget):
    pui_terminal = False
    def __init__(self, editable=False, index_model=None, text_model=None):
        super().__init__()
        self.editable = editable
        self.index_model = index_model
        self.text_model = text_model

    def update(self, prev):
        if prev and prev.ui:
            self.curr_index = prev.curr_index
            self.curr_text = prev.curr_text
            self.ui = prev.ui
        else:
            self.curr_index = Prop()
            self.curr_text = Prop()
            self.ui = wx.ComboBox(getWindow(self.parent), choices=[], style=0 if self.editable else wx.CB_READONLY)
            self.ui.Bind(wx.EVT_COMBOBOX, self._combobox)
            self.ui.Bind(wx.EVT_TEXT, self._text)

        super().update(prev)

    def postSync(self):
        index = -1
        text = ""

        if self.index_model:
            index = self.index_model.value
            text = self.children[index].text
        elif self.text_model:
            text = str(self.text_model.value)
            try:
                index = [c.value for c in self.children].index(text)
            except:
                index = -1

        if self.curr_index.set(index):
            self.ui.SetSelection(index)

        if self.curr_text.set(text):
            self.ui.SetValue(text)

    def _combobox(self, *args, **kwargs):
        node = self.get_node()
        idx = self.ui.GetSelection()
        if node.index_model:
            node.index_model.value = idx
        if node.text_model:
            node.text_model.value = self.children[idx].value
        e = PUIEvent()
        e.value = idx
        node._change(e)

    def _text(self, *args, **kwargs):
        if not self.editable:
            return
        node = self.get_node()
        text = self.ui.GetValue()
        if node.text_model:
            node.text_model.value = text
        e = PUIEvent()
        e.value = text
        node._change(e)

    def addChild(self, idx, child):
        self.ui.Insert(child.text, idx)

    def removeChild(self, idx, child):
        self.ui.Delete(idx)

class ComboBoxItem(PUINode):
    def __init__(self, text, value=None):
        super().__init__()
        self.id(text)
        self.text = text
        if value is None:
            value = text
        self.value = value



from .view import *

def PUI(func):
    """
    PUI.PUI triggers update() directly by redraw()
    """
    def func_wrapper(*args, **kwargs):
        class PUIViewWrapper(PUIView):
            pui_virtual = True
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper


from enum import Enum
from enum import IntEnum
class Anchor(Enum):
    LEFT_TOP = ("left" ,"top")
    LEFT_CENTER = ("left" ,"center")
    LEFT_BOTTOM = ("left" ,"bottom")
    CENTER_TOP = ("center" ,"top")
    CENTER = ("center" ,"center")
    CENTER_BOTTOM = ("center" ,"bottom")
    RIGHT_TOP = ("right" ,"top")
    RIGHT_CENTER = ("right" ,"center")
    RIGHT_BOTTOM = ("right" ,"bottom")

class MouseButton(IntEnum):
    NONE = 0
    LEFT = 1
    RIGHT = 2
    MIDDLE = 4
    X1 = 8
    X2 = 16

class KeyModifier(IntEnum):
    SHIFT = 1
    CTRL = 2
    ALT = 4
    META = 8



from .. import *
from .base import *

class Application(PUIView):
    def __init__(self):
        super().__init__()
        self.ready = False

    def sync(self):
        if not self.ready:
            return
        super().sync()

    def flet_app(self, page: ft.Page):
        self.ui = page
        self.ready = True
        self.sync()

    def addChild(self, idx, child):
        if idx > 0:
            print("Flet backend only support single window")

    def removeChild(self, idx, child):
        pass

    def start(self):
        ft.app(self.flet_app)

def PUIApp(func):
    def func_wrapper(*args, **kwargs):
        class PUIAppWrapper(Application):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIAppWrapper(func.__name__)
        return ret

    return func_wrapper



from .. import *
from .base import *
from .layout import *

class Divider(WxBaseWidget):
    def __init__(self):
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            if isinstance(self.non_virtual_parent, VBox):
                style = wx.LI_HORIZONTAL
            else:
                style = wx.LI_VERTICAL
            self.ui = wx.StaticLine(getWindow(self.parent), style=style, size=(15, 15))

        super().update(prev)



from .. import *
from .base import *

class ProgressBar(WxBaseWidget):
    expand_x_prio = 1

    def __init__(self, progress, maximum=1):
        super().__init__()
        self.progress = progress
        self.maximum = maximum

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = wx.Gauge(getWindow(self.parent))
        self.ui.SetRange(self.maximum)
        self.ui.SetValue(self.progress)
        super().update(prev)



DEBUG = False

class VDomError(Exception):
    pass

def recur_delete(node, child, direct):
    child.destroyed = True
    for sc in child.children:
        if sc is not None:
            recur_delete(child, sc, False)
    if DEBUG:
        print("recur_delete", node.key, child.key, direct)
    child.destroy(direct)

def sortGridDOMInPlace(dom):
    dom[:] = [c for c in dom if c.grid_row is not None and c.grid_column is not None]
    dom.sort(key=lambda c:(c.grid_row, c.grid_column, c.grid_rowspan, c.grid_columnspan))

def dom_remove_node(dom_parent, dom_offset, child):
    if DEBUG:
        print(f"dom_remove_node key={child.key} virtual={child.pui_virtual} children={len(child.children)} dom_parent={dom_parent.key} dom_offset={dom_offset}")
    if child.pui_virtual:
        ret = [child]
        for c in child.children:
            ret.extend(dom_remove_node(dom_parent, dom_offset, c))
        return ret
    else:
        dom_parent.removeChild(dom_offset , child)
        return [child]

def dom_add_nodes(dom_parent, dom_offset, children):
    for childIdx, c in enumerate([n for n in children if not n.pui_virtual]):
        dom_parent.addChild(dom_offset+childIdx, c)

def countDomChildren(nodes):
    num = 0
    for c in nodes:
        if c.pui_virtual:
            num += countDomChildren(c.children)
        elif not c.pui_outoforder:
            num += 1
    return num

def sync(node, dom_parent, dom_offset, oldVDOM, newVDOM, depth=0):
    orig_dom_children_num = dom_children_num = countDomChildren(oldVDOM)
    dom_children_curr = 0

    if DEBUG:
        print(f"{(depth)*'    '}Syncing {node.key}@{id(node)}#tag={node._tag} parent={dom_parent.key}@{id(dom_parent)}#tag={dom_parent._tag} dom_offset={dom_offset} dom_children_num={dom_children_num} old={len(oldVDOM)} new={len(newVDOM)}")

    if node.pui_grid_layout:
        sortGridDOMInPlace(oldVDOM)
        sortGridDOMInPlace(newVDOM)
    else:
        oldOrdered = [c for c in oldVDOM if not c.pui_outoforder]
        oldOutOfOrdered = [c for c in oldVDOM if c.pui_outoforder]
        newOrdered = [c for c in newVDOM if not c.pui_outoforder]
        newOutOfOrdered = [c for c in newVDOM if c.pui_outoforder]

        if node.pui_reversed_order:
            oldOrdered.reverse()
            newOrdered.reverse()

        oldVDOM.clear()
        oldVDOM.extend(oldOrdered + oldOutOfOrdered)
        newVDOM.clear()
        newVDOM.extend(newOrdered + newOutOfOrdered)

    if DEBUG:
        print(f"{(depth+1)*'    '}===OLD===")
        for c in oldVDOM:
            print(f"{(depth+1)*'    '}{c.key}#tag={c._tag} virtual={c.pui_virtual} children={len(c.children)} ui={c.ui}")

        print(f"{(depth+1)*'    '}===NEW===")
        for c in newVDOM:
            print(f"{(depth+1)*'    '}{c.key}#tag={c._tag} virtual={c.pui_virtual} children={len(c.children)}")
        print(f"{(depth+1)*'    '}=========")

    oldVMap = [x.key for x in oldVDOM]
    newVMap = [x.key for x in newVDOM]

    node.preSync()

    toBeDeleted = []
    for childIdx, new in enumerate(newVDOM):
        if DEBUG:
            print(f"{(depth+1)*'    '}sync child {childIdx}, {new.key} dom_parent={dom_parent.key} virtual={new.pui_virtual}")
        new.pui_dom_parent = dom_parent

        while True:
            # Step 1. just matched
            if childIdx < len(oldVDOM) and oldVMap[childIdx] == new.key: # matched
                if DEBUG:
                    print(f"{(depth+1)*'    '}S1. MATCHED {childIdx} {new.key}")
                old = oldVDOM[childIdx]

                if old.pui_isview: # must also be virtual
                    n = countDomChildren(old.children)
                    dom_children_curr += n
                    if DEBUG:
                        print(f"{(depth+1)*'    '}    dom_children_curr += {n} => {dom_children_curr} dom_children_num={dom_children_num}")

                    old.parent = new.parent
                    old.pui_dom_parent = new.pui_dom_parent
                    newVDOM[childIdx] = old
                    new.destroy(True) # deregister old view from PUIView.__ALLVIEWS__
                else:
                    try:
                        new.update(old)
                    except:
                        import traceback
                        print("## <ERROR OF update() >")
                        print(new.key)
                        traceback.print_exc()
                        print("## </ERROR OF update()>")

                    if new.pui_virtual:
                        num, delta = sync(new, node, dom_offset + dom_children_curr, old.children, new.children, depth+1)
                        dom_children_curr += num
                        dom_children_num += delta
                        if DEBUG:
                            print(f"{(depth+2)*'    '}dom_children_curr += {num} => {dom_children_curr} dom_children_num += {delta} => {dom_children_num}")
                    else:
                        if not new.pui_outoforder:
                            dom_children_curr += 1
                            if DEBUG:
                                print(f"{(depth+2)*'    '}dom_children_curr += 1 => {dom_children_curr} dom_children_num={dom_children_num}")

                        if not new.pui_terminal:
                            sync(new, old, 0, old.children, new.children, depth+1)

                break # finish

            # Step 2. trim removed nodes after common prefix
            trimmed = False
            while childIdx < len(oldVDOM) and not oldVDOM[childIdx].key in newVMap[childIdx:]: # trim old nodes
                if DEBUG:
                    print(f"{(depth+1)*'    '}S2. TRIM {childIdx} {oldVDOM[childIdx].key}")
                old = oldVDOM.pop(childIdx)
                oldVMap.pop(childIdx)
                nodes = dom_remove_node(dom_parent, dom_offset + dom_children_curr, old)
                n = len([n for n in nodes if not n.pui_virtual and not n.pui_outoforder])
                dom_children_num -= n
                if DEBUG:
                    print(f"{(depth+2)*'    '}dom_children_num -= {n} => {dom_children_num}")
                toBeDeleted.extend(nodes)
                trimmed = True

            if trimmed:
                continue # restart

            # Step 3. setup target node

            matchedIdx = None
            if new.pui_movable:
                try:
                    matchedIdx = oldVMap[childIdx+1:].index(new.key) + childIdx + 1
                except ValueError:
                    pass

            ## Step 3-1. new node
            if matchedIdx is None:
                if DEBUG:
                    print(f"{(depth+1)*'    '}S3-1. NEW {childIdx} {new.key}")
                # always populate new node regardless of isview or not
                try:
                    new.update(None)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(new.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")

                if new.pui_virtual:
                    num, delta = sync(new, dom_parent, dom_offset + dom_children_curr, [], new.children, depth+1)
                    dom_children_curr += num
                    dom_children_num += num
                    if DEBUG:
                        print(f"{(depth+2)*'    '}dom_children_curr += {num} => {dom_children_curr} dom_children_num += {num} => {dom_children_num}")
                else:
                    if DEBUG:
                        print(f"{(depth+1)*'    '}addChild", dom_parent.key, dom_offset + dom_children_curr, new.key)
                    dom_parent.addChild(dom_offset + dom_children_curr, new)

                    if not new.pui_outoforder:
                        dom_children_curr += 1
                        dom_children_num += 1
                        if DEBUG:
                            print(f"{(depth+2)*'    '}dom_children_curr += 1 => {dom_children_curr} dom_children_num += 1 => {dom_children_num}")

                    if not new.pui_terminal:
                        sync(new, new, 0, [], new.children, depth+1)

                oldVDOM.insert(childIdx, new) # put new node back for later findDomOffsetForNode
                oldVMap.insert(childIdx, new.key)

            ## Step 3-2. existed node
            else:
                # if the target node is in just next position, requeue the blocker to prevent repositioning every nodes coming after
                # eg. when an element is removed from a long list, do single 3-2-1 instead of many 3-2-2

                ### Step 3-2-1. yield the next position for the target node
                if matchedIdx == childIdx + 1:
                    if DEBUG:
                        print(f"{(depth+1)*'    '}S3-2-1. YIELD {childIdx} {new.key}")
                    oldVMap.pop(childIdx)
                    toBeRequeued = oldVDOM.pop(childIdx)
                    nodes = dom_remove_node(dom_parent, dom_offset + dom_children_curr, toBeRequeued)
                    n = len([n for n in nodes if not n.pui_virtual and not n.pui_outoforder])
                    dom_add_nodes(dom_parent, dom_children_num - n, nodes)
                    oldVMap.append(toBeRequeued.key)
                    oldVDOM.append(toBeRequeued)

                ### Step 3-2-2. move target node
                else:
                    if DEBUG:
                        print(f"{(depth+1)*'    '}S3-2-2. MOVE {childIdx} {new.key}")
                    found, offset = dom_parent.findDomOffsetForNode(old)
                    if not found:
                        raise VDomError(f"S3-2-2: findDomOffsetForNode() failed for {old.key}#tag={old._tag} on {dom_parent.key}#tag={dom_parent._tag} {dom_parent.children}")
                    oldVMap.pop(matchedIdx)
                    old = oldVDOM.pop(matchedIdx)
                    nodes = dom_remove_node(dom_parent, offset, old)
                    dom_add_nodes(dom_parent, dom_offset + dom_children_curr, nodes)

                    oldVDOM.insert(childIdx, new) # put new node back for later findDomOffsetForNode
                    oldVMap.insert(childIdx, new.key)

                continue # restart, sync will be peformed in next step 1

            break # finish

    # Step 4. trim removed trail
    if DEBUG:
        print(f"{(depth+1)*'    '}S4. TRIM")
    nl = len(newVDOM)
    if DEBUG:
        print(f"{(depth+1)*'    '}S4. TRIM", f"dom_offset={dom_offset}", len(oldVDOM), "=>", len(newVDOM))
    while len(oldVDOM) > nl:
        old = oldVDOM.pop(nl)
        if DEBUG:
            print(f"{(depth+2)*'    '}", f"key={old.key} virtual={old.pui_virtual} children={len(old.children)}")
        oldVMap.pop(nl)
        nodes = dom_remove_node(dom_parent, dom_offset + nl, old)
        dom_children_num -= len([n for n in nodes if not n.pui_virtual and not n.pui_outoforder])
        toBeDeleted.append(old)

    for c in newVDOM:
        c.postUpdate()

    node.postSync()

    # release deleted nodes
    for old in toBeDeleted:
        recur_delete(node, old, True)

    if DEBUG:
        print(f"{(depth)*'    '}sync end {node.key} -> {dom_children_curr},{dom_children_num}")

    if dom_children_curr != dom_children_num:
        raise VDomError(f"dom_children_curr != dom_children_num for {node.key} -> {dom_children_curr},{dom_children_num}")
    return dom_children_curr, dom_children_num - orig_dom_children_num


from .. import *
from .base import *

class Application(TPUIView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.theme = None

    def tkinter(self, theme=None, **kwargs):
        super().tkinter(**kwargs)
        self.theme = theme
        return self

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = tk.Tk()
            ttkStyle = ttk.Style(self.ui)
            if self.theme:
                ttkStyle.theme_use(self.theme)
            self.ui.withdraw()

        super().update(prev)

    def addChild(self, idx, child):
        pass

    def removeChild(self, idx, child):
        pass

    def start(self):
        self.ui.mainloop()


def PUIApp(func):
    def func_wrapper(*args, **kwargs):
        class PUIAppWrapper(Application):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args, **kwargs)

        ret = PUIAppWrapper(func.__name__)
        return ret

    return func_wrapper



from .. import *
from .base import *

class RadioButton(WxBaseWidget):
    def __init__(self, text, value, model):
        super().__init__()
        self.var = None
        self.text = text
        self.value = value
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.SetLabel(self.text)
        else:
            self.ui = wx.RadioButton(getWindow(self.parent), label=self.text, style=wx.RB_SINGLE)
            self.ui.Bind(wx.EVT_RADIOBUTTON, self._selected)

        self.ui.SetValue(self.model.value == self.value)

        super().update(prev)

    def _selected(self, *args, **kwargs):
        node = self.get_node()
        node.model.value = node.value
        node._clicked()


from .. import *
from .base import *

class Stack(QtBaseLayout):
    pui_terminal = False
    pui_reversed_order = True

    def __init__(self):
        super().__init__()
        self.qt_params = {}
        if not isinstance(self.non_virtual_parent, QtBaseLayout):
            self.layout_padding = (11,11,11,11)

    @property
    def outer(self):
        return self.ui

    @property
    def inner(self):
        return self.layout

    def destroy(self, direct):
        self.layout = None
        super().destroy(direct)

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
            self.qtlayout = prev.qtlayout
        else:
            self.ui = QtWidgets.QWidget()
            self.qtlayout = QtWidgets.QStackedLayout()
            self.qtlayout.setStackingMode(QtWidgets.QStackedLayout.StackAll)
            self.qtlayout.setContentsMargins(0,0,0,0)
            self.ui.setLayout(self.qtlayout)
        super().update(prev)

    def addChild(self, idx, child):
        from .modal import Modal
        from .layout import Spacer
        if isinstance(child, Spacer):
            pass
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.qtlayout.insertWidget(idx, child.outer)
            self.mounted_children.insert(idx, child)

    def removeChild(self, idx, child):
        from .modal import Modal
        from .layout import Spacer
        if isinstance(child, Spacer):
            pass
        elif isinstance(child, Modal):
            pass
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
            self.mounted_children.pop(idx)

    def postUpdate(self):
        if self.ui:
            if self._onDropped:
                self.ui.setAcceptDrops(True)
                self.ui.installEventFilter(self.eventFilter)
            else:
                self.ui.setAcceptDrops(False)

        super().postUpdate()

class HBox(QtBaseLayout):
    container_x = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.qtlayout = prev.qtlayout
        else:
            self.ui = QtWidgets.QWidget()
            self.qtlayout = QtWidgets.QHBoxLayout()
            self.qtlayout.setContentsMargins(0,0,0,0)
            self.ui.setLayout(self.qtlayout)
        super().update(prev)

class VBox(QtBaseLayout):
    container_y = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.qtlayout = prev.qtlayout
        else:
            self.ui = QtWidgets.QWidget()
            self.qtlayout = QtWidgets.QVBoxLayout()
            self.qtlayout.setContentsMargins(0,0,0,0)
            self.ui.setLayout(self.qtlayout)
        super().update(prev)

class Spacer(PUINode):
    pui_terminal = True
    pui_movable = False

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            if isinstance(self.non_virtual_parent, VBox):
                self.ui = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
            elif isinstance(self.non_virtual_parent, HBox):
                self.ui = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
            else:
                self.ui = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        super().update(prev)

    def destroy(self, direct):
        # self.ui.deleteLater() # QSpacerItem doesn't have .deleteLater()
        self.ui = None
        super().destroy(direct)

class Grid(QtBaseLayout):
    pui_grid_layout = True

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.layout = prev.layout
        else:
            self.ui = QtWidgets.QWidget()
            self.layout = QtWidgets.QGridLayout()
            self.layout.setContentsMargins(0,0,0,0)
            self.ui.setLayout(self.layout)
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.layout.addWidget(child.outer, child.grid_row, child.grid_column, child.grid_rowspan or 1, child.grid_columnspan or 1)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])


from .. import *
from .base import *
from ..utils import *
from PySide6.QtWidgets import QSizePolicy

class Text(QtBaseWidget):
    textformat = QtCore.Qt.TextFormat.PlainText
    def __init__(self, text, selectable=False):
        super().__init__()
        self.text = str(text)
        self.selectable = selectable

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.setText(self.text)
        else:
            self.ui = QtWidgets.QLabel(self.text)
            self.ui.setTextFormat(self.textformat)
            self.ui.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
            self.ui.setWordWrap(True)

        if self.selectable:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        else:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.NoTextInteraction)

        super().update(prev)


class Html(Text):
    textformat = QtCore.Qt.TextFormat.RichText

class MarkDown(Text):
    textformat = QtCore.Qt.TextFormat.MarkdownText



from .. import *
from .base import *
from ..utils import *
from PySide6.QtWidgets import QSizePolicy

class RadioButton(QtBaseWidget):
    def __init__(self, text, value, model):
        super().__init__()
        self.text = text
        self.value = value
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            try:
                self.ui.clicked.disconnect()
            except:
                pass
        else:
            self.ui = QtWidgets.QRadioButton()
        self.ui.setText(self.text)
        self.ui.setChecked(self.model.value == self.value)
        self.ui.clicked.connect(self._clicked)
        super().update(prev)

    def _clicked(self):
        self.model.value = self.value
        super()._clicked()



from .. import *
from .base import *
from .menu import *

class QtModal(QtWidgets.QDialog):
    def closeEvent(self, arg__1) -> None:
        self.puinode._close()
        return super().closeEvent(arg__1)

class Modal(QtBaseWidget):
    pui_terminal = False
    pui_outoforder = True

    def __init__(self, status, offValue=None, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.status = status
        self.offValue = offValue
        self.title = title
        self.size = size
        self.maximize = maximize
        self.fullscreen = fullscreen
        self.open_cb = None
        self.close_cb = None

    @property
    def outer(self):
        return self.ui

    @property
    def inner(self):
        return self.layout

    def destroy(self, direct):
        self.close_modal()
        return super().destroy(direct)

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
            self.curr_status = prev.curr_status
        else:
            self.ui = QtModal()
            self.ui.setModal(True)
            self.layout = QtWidgets.QVBoxLayout()
            self.ui.setLayout(self.layout)
            self.layout.setContentsMargins(0,0,0,0)
            self.curr_size = Prop()
            self.curr_maximize = Prop()
            self.curr_fullscreen = Prop()
            self.curr_status = Prop()

        if self.curr_size.set(self.size):
            self.ui.resize(*self.size)
        if self.curr_maximize.set(self.maximize):
            self.ui.showMaximized()
        if self.curr_fullscreen.set(self.fullscreen):
            self.ui.showFullScreen()
        if not self.title is None:
            self.ui.setWindowTitle(self.title)

        self.ui.puinode = self

        if self.status.value:
            self.open_modal()
        else:
            self.close_modal()
        super().update(prev)

    def open(self, cb, *args, **kwargs):
        self.open_cb = (cb, args, kwargs)
        return self

    def close(self, cb, *args, **kwargs):
        self.close_cb = (cb, args, kwargs)
        return self

    def open_modal(self):
        if not self.curr_status:
            self.ui.show()
            self.curr_status = True
            if self.open_cb:
                self.open_cb[0](*self.open_cb[1], **self.open_cb[2])

    def close_modal(self):
        prev_status = self.curr_status
        if self.curr_status is None or self.curr_status:
            self.curr_status = False
        if prev_status:
            self.ui.close()
            if self.close_cb:
                self.close_cb[0](*self.close_cb[1], **self.close_cb[2])

    def _close(self, *args, **kwargs):
        node = self.get_node()
        if node.curr_status:
            node.curr_status = False
            if node.close_cb:
                node.close_cb[0](*node.close_cb[1], **node.close_cb[2])
        node.status.value = node.offValue


    def addChild(self, idx, child):
        from .layout import Spacer
        if isinstance(child, MenuBar):
            self.ui.setMenuBar(child.outer)
        elif isinstance(child, Spacer):
            self.layout.insertItem(idx, child.outer)
        elif isinstance(child, Modal):
            child.outer.show()
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            params = {}
            if not child.layout_weight is None:
                params["stretch"] = child.layout_weight
            self.layout.insertWidget(idx, child.outer, **params)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        from .layout import Spacer
        if isinstance(child, MenuBar):
            child.outer.close()
        elif isinstance(child, Spacer):
            self.layout.removeItem(child.outer)
        elif isinstance(child, Modal):
            child.outer.close()
        elif isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])



from .. import *
from .base import *

class Window(FBase):
    def __init__(self, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.size = size
        self.maximize = maximize
        self.fullscreen = fullscreen
        self.child_weight = 1

    def update(self, prev=None):
        self.inner.title = self.title
        super().update(prev)

    def addChild(self, idx, child):
        if idx != 0:
            return
        self.inner.add(child.outer)

    def removeChild(self, idx, child):
        if idx != 0:
            return
        self.inner.remove(child.outer)



from .. import *
from .base import *

class Text(WxBaseWidget):
    def __init__(self, text, selectable=False):
        super().__init__()
        self.text = text
        self.selectable = selectable

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.SetLabel(self.text)
        else:
            self.ui = wx.StaticText(getWindow(self.parent), label=self.text)

        super().update(prev)

class Html(Text):
    pui_supported = False

class MarkDown(Text):
    pui_supported = False



from .. import *
from .base import *
import wx.lib.scrolledpanel as scrolled

class Scroll(WxBaseWidget):
    pui_terminal = False
    scroll = True

    END = -0.0

    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = scrolled.ScrolledPanel(getWindow(self.parent))

        self.container_y = True
        self.expand_y_prio = 3
        if self.vertical is False:
            self.container_y = False
            self.expand_y_prio = 1

        self.container_x = True
        self.expand_x_prio = 3
        if self.horizontal is False:
            self.container_x = False
            self.expand_x_prio = 1

        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            self.ui.SetSizer(child.outer)
        elif isinstance(child, WxBaseWidget):
            pass

    def removeChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            pass
        elif isinstance(child, WxBaseWidget):
            pass

    def scrollX(self, pos=0):
        return self

    def scrollY(self, pos=0):
        return self

    def postSync(self):
        self.ui.SetupScrolling(scroll_x=self.horizontal or self.horizontal is None, scroll_y=self.vertical or self.vertical is None, scrollToTop=False)



from .. import *
from .base import *
import math

class Canvas(WxBaseWidget):
    expand_x_prio = 3
    expand_y_prio = 3

    def __init__(self, painter, *args):
        super().__init__()
        self.ui = None
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.Refresh()
        else:
            self.ui = wx.Panel(getWindow(self.parent))
            self.ui.Bind(wx.EVT_PAINT, self._paint)
            self.ui.Bind(wx.EVT_LEFT_DCLICK, self._LeftDblClick)
            self.ui.Bind(wx.EVT_LEFT_DOWN, self._LeftDown)
            self.ui.Bind(wx.EVT_LEFT_UP, self._LeftUp)
            self.ui.Bind(wx.EVT_MOTION, self._Motion)
            self.ui.Bind(wx.EVT_MOUSEWHEEL, self._MouseWheel)
            self.ui.SetMinSize((self.layout_width, self.layout_height))

        super().update(prev)

    def _MouseWheel(self, event):
        e = PUIEvent()
        e.x, e.y = event.GetPosition()
        if event.GetWheelAxis() == wx.MOUSE_WHEEL_VERTICAL:
            e.h_delta = 0
            e.v_delta = event.GetWheelRotation()
            e.x_delta = 0
            e.y_delta = event.GetWheelDelta()
        else:
            e.h_delta = event.GetWheelRotation()
            e.v_delta = 0
            e.x_delta = event.GetWheelDelta()
            e.y_delta = 0
        self.get_node()._wheel(e)

    def _LeftDblClick(self, event):
        e = PUIEvent()
        e.x, e.y = event.GetPosition()
        self.get_node()._dblclicked(e)

    def _LeftDown(self, event):
        e = PUIEvent()
        e.x, e.y = event.GetPosition()
        self.get_node()._mousedown(e)

    def _LeftUp(self, event):
        e = PUIEvent()
        e.x, e.y = event.GetPosition()
        self.get_node()._mouseup(e)

    def _Motion(self, event):
        e = PUIEvent()
        e.x, e.y = event.GetPosition()
        self.get_node()._mousemove(e)

    def _paint(self, event):
        node = self.get_node()

        node.dc = wx.PaintDC(node.ui)
        if not node.style_bgcolor is None:
            original_brush = node.dc.GetBrush()
            node.dc.SetBrush(wx.Brush(int_to_wx_colour(node.style_bgcolor)))
            node.dc.DrawRectangle(self.ui.GetClientRect())
            node.dc.SetBrush(original_brush)

        node.painter(node, *node.args)
        node.dc = None


    def drawText(self, x, y, text, w=None, h=None, size=12, color=None, rotate=0, anchor=Anchor.LEFT_TOP):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        lines = text.split("\n")
        extents = [self.dc.GetTextExtent(l) for l in lines]
        width = max([e[0] for e in extents])
        height = sum([e[1] for e in extents])

        dx = 0
        dy = 0
        if anchor.value[0]=="center":
            dx = width/2
        elif anchor.value[0]=="right":
            dx = width

        if anchor.value[1]=="center":
            dy = height/2
        elif anchor.value[1]=="bottom":
            dy = height

        gc = wx.GraphicsContext.Create(self.dc)

        gc.PushState()

        gc.Translate(x+dx, y+dy)
        gc.Rotate(math.radians(rotate))
        gc.Translate(-dx, -dy)

        if color is None:
            color = self.ui.GetForegroundColour()
        else:
            color = int_to_wx_colour(color)

        gc.SetFont(wx.Font(size, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL), color)
        y_off = 0
        for i,line in enumerate(lines):
            gc.PushState()
            gc.DrawText(line, 0, y_off)
            gc.PopState()
            y_off += extents[i][1]

        gc.PopState()

        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawLine(self, x1, y1, x2, y2, color=0, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        self.dc.SetPen(wx.Pen(int_to_wx_colour(color), width))
        self.dc.DrawLine(int(x1), int(y1), int(x2), int(y2))
        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawPolyline(self, coords, color=0, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        if color is not None:
            self.dc.SetPen(wx.Pen(int_to_wx_colour(color), width))

        coords = [(int(c[0]), int(c[1])) for c in coords]
        self.dc.DrawLines(coords)

        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawPolygon(self, coords, fill=None, stroke=None, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        if fill is None:
            self.dc.SetBrush(wx.TRANSPARENT_BRUSH)
        else:
            self.dc.SetBrush(wx.Brush(int_to_wx_colour(fill)))
        if stroke is None:
            self.dc.SetPen(wx.NullPen)
        else:
            self.dc.SetPen(wx.Pen(int_to_wx_colour(stroke), width))

        coords = [(int(c[0]), int(c[1])) for c in coords]
        self.dc.DrawPolygon(coords)
        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawRect(self, x1, y1, x2, y2, fill=None, stroke=None, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        if fill is None:
            self.dc.SetBrush(wx.TRANSPARENT_BRUSH)
        else:
            self.dc.SetBrush(wx.Brush(int_to_wx_colour(fill)))
        if stroke is None:
            self.dc.SetPen(wx.NullPen)
        else:
            self.dc.SetPen(wx.Pen(int_to_wx_colour(stroke), width))

        x = min(x1, x1)
        y = min(y1, y2)
        w = abs(x2 - x1)
        h = abs(y2 - y1)

        self.dc.DrawRectangle(int(x), int(y), int(w), int(h))
        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawEllipse(self, x, y, rx, ry, fill=None, stroke=None, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        if fill is None:
            self.dc.SetBrush(wx.TRANSPARENT_BRUSH)
        else:
            self.dc.SetBrush(wx.Brush(int_to_wx_colour(fill)))
        if stroke is None:
            self.dc.SetPen(wx.NullPen)
        else:
            self.dc.SetPen(wx.Pen(int_to_wx_colour(stroke), width))

        self.dc.DrawEllipse(int(x-rx), int(y-ry), int(rx*2), int(ry*2))
        self.dc.SetPen(original_pen)
        self.dc.SetBrush(original_brush)

    def drawShapely(self, shape, fill=None, stroke=None, width=1):
        original_pen = self.dc.GetPen()
        original_brush = self.dc.GetBrush()

        if fill is None:
            self.dc.SetBrush(wx.TRANSPARENT_BRUSH)
        else:
            self.dc.SetBrush(wx.Brush(int_to_wx_colour(fill)))
        if stroke is None:
            self.dc.SetPen(wx.NullPen)
        else:
            self.dc.SetPen(wx.Pen(int_to_wx_colour(stroke), width))

        self._drawShapely(shape, fill, stroke, width)

    def _drawShapely(self, shape, fill=None, stroke=None, width=1):
        if hasattr(shape, "geoms"):
            for g in shape.geoms:
                self.drawShapely(g, fill, stroke, width)
        elif hasattr(shape, "exterior"): # polygon
            gc = wx.GraphicsContext.Create(self.dc)
            if gc:
                path = gc.CreatePath()

                exterior = shape.exterior.coords
                path.MoveToPoint(* exterior[0])
                for point in exterior[1:]:
                    path.AddLineToPoint(*point)
                path.CloseSubpath()

                for hole in shape.interiors:
                    hole = hole.coords
                    path.MoveToPoint(*hole[0])
                    # Draw holes in reverse order
                    for point in hole[1:]:
                        path.AddLineToPoint(*point)
                    path.CloseSubpath()

                if fill is not None:
                    gc.SetBrush(wx.Brush(int_to_wx_colour(fill)))
                if stroke is not None:
                    gc.SetPen(wx.Pen(int_to_wx_colour(stroke), width))

                gc.DrawPath(path)
        elif hasattr(shape, "x") and hasattr(shape, "y"): # point
            self.drawEllipse(shape.x, shape.y, width/2, width/2, fill=stroke)
        elif hasattr(shape, "coords"): # linestring, linearring
            self.drawPolyline(shape.coords, color=stroke, width=width)
        else:
            raise RuntimeError(f"Not implemented: drawShapely({type(shape).__name__}) {dir(shape)}")


from .. import *
from .base import *

class Checkbox(TBase):
    def __init__(self, text, model):
        super().__init__()
        self.text = text
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.model.value
        else:
            self.ui = widgets.Checkbox(self.text, self.model.value)

        self.ui.puinode = self
        super().update(prev)

    def _changed(self, value):
        self.model.value = value



from .. import *

import wx

def int_to_wx_colour(color_int):
    red = (color_int >> 16) & 0xFF
    green = (color_int >> 8) & 0xFF
    blue = color_int & 0xFF
    return wx.Colour(red, green, blue)

def getWindow(n):
    from .window import Window
    from .scroll import Scroll
    p = n
    while p:
        if isinstance(p, Scroll):
            return p.ui
        if isinstance(p, Window):
            return p.ui
        p = p.parent
    raise RuntimeError("!!! getWindow returns None for", n.key if n else None)

class WxPUIView(PUIView):
    pui_virtual = True
    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        if self.ui: # PUIView doesn't have ui
            self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def redraw(self):
        wx.CallAfter(self.sync)

    def update(self, prev=None):
        if self.retired_by:
            return
        self.dirty = False
        super().update(prev)
        self.updating = False
        if self.dirty:
            self.update(prev)

class WXBase(PUINode):
    scroll = False
    container_x = False # axis
    container_y = False # axis
    expand_x_prio = 0
    expand_y_prio = 0
    expand_x1_children = 0
    expand_x2_children = 0
    expand_x3_children = 0
    expand_x4_children = 0
    expand_y1_children = 0
    expand_y2_children = 0
    expand_y3_children = 0
    expand_y4_children = 0
    cached_wxparent = None

    @property
    def expand_x(self):
        return self.expand_x_prio

    @property
    def expand_y(self):
        return self.expand_y_prio

    def update(self, prev):
        self.cached_wxparent = parent = self.wxparent

        parent = self.wxparent
        if parent:
            if self.layout_weight:
                if parent.container_x:
                    self.expand_x_prio = 4
                if parent.container_y:
                    self.expand_y_prio = 4

            if self.expand_x_prio >= 1:
                parent.expand_x1_children += 1
            if self.expand_x_prio >= 2:
                parent.expand_x2_children += 1
            if self.expand_x_prio >= 3:
                parent.expand_x3_children += 1
            if self.expand_x_prio >= 4:
                parent.expand_x4_children += 1

            if self.expand_y_prio >= 1:
                parent.expand_y1_children += 1
            if self.expand_y_prio >= 2:
                parent.expand_y2_children += 1
            if self.expand_y_prio >= 3:
                parent.expand_y3_children += 1
            if self.expand_y_prio >= 4:
                parent.expand_y4_children += 1

        super().update(prev)

    def postUpdate(self):
        parent = self.cached_wxparent
        if parent:
            if parent.container_x:
                if self.expand_x_prio < 1 and parent.expand_x1_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 2 and parent.expand_x2_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 3 and parent.expand_x3_children > 0:
                    self.expand_x_prio = 0
                if self.expand_x_prio < 4 and parent.expand_x4_children > 0:
                    self.expand_x_prio = 0

            if parent.container_y:
                if self.expand_y_prio < 1 and parent.expand_y1_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 2 and parent.expand_y2_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 3 and parent.expand_y3_children > 0:
                    self.expand_y_prio = 0
                if self.expand_y_prio < 4 and parent.expand_y4_children > 0:
                    self.expand_y_prio = 0

        if self._debug:
            print("layout", self.key, f"weight={self.layout_weight}", f"expand_x={self.expand_x}", f"expand_y={self.expand_y}")

        super().postUpdate()

    @property
    def wxparent(self):
        parent = self.parent
        while not isinstance(parent, WXBase):
            if parent==parent.parent:
                parent = None
                break
            parent = parent.parent
        return parent


class WxBaseWidget(WXBase):
    pui_terminal = True

    def __init__(self):
        super().__init__()
        self.layout_padding = (3, 3, 3, 3)

    def update(self, prev):
        super().update(prev)

        if self.style_fontsize:
            font = self.ui.GetFont()
            font.SetPointSize(self.style_fontsize)
            self.ui.SetFont(font)

        # Set font color if specified
        if self.style_color:
            self.ui.SetForegroundColour(int_to_wx_colour(self.style_color))

    def destroy(self, direct):
        if self.ui is not None:
            self.ui.Destroy()
            self.ui = None
        super().destroy(direct)

class WxBaseLayout(WXBase):
    def __init__(self):
        super().__init__()
        self.relayout = False
        if not isinstance(self.non_virtual_parent, WxBaseLayout):
            self.layout_padding = (11,11,11,11)

    def destroy(self, direct):
        # self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def update(self, prev):
        if prev and prev.ui:
            self.sizerItems = prev.sizerItems
        else:
            self.sizerItems = []
        super().update(prev)

    def postUpdate(self):
        super().postUpdate()

        for child, si in self.sizerItems:
            if child.expand_x or child.expand_y:
                si.SetFlag(wx.ALL | wx.EXPAND)
            else:
                # si.SetFlag(wx.ALL | wx.ALIGN_CENTER)
                si.SetFlag(wx.ALL) # XXX

            weight = child.layout_weight
            if self.container_x and child.expand_x:
                si.SetProportion(weight if weight else 1 if child.expand_x else 0)
            elif self.container_y and child.expand_y:
                si.SetProportion(weight if weight else 1 if child.expand_y else 0)

            p = 0
            if child.layout_padding:
                p = max(child.layout_padding)
            si.SetBorder(p)

    def addChild(self, idx, child):
        from .layout import Spacer
        self.relayout = True

        weight = child.layout_weight
        if weight is None:
            weight = 0
        if not weight and self.container_x and child.expand_x:
            weight = 1
        if not weight and self.container_y and child.expand_y:
            weight = 1

        if isinstance(child, WxBaseLayout):
            si = self.ui.Insert(idx, child.outer)
            self.sizerItems.insert(idx, (child, si))
        elif isinstance(child, WxBaseWidget):
            si = self.ui.Insert(idx, child.outer)
            self.sizerItems.insert(idx, (child, si))
        elif isinstance(child, Spacer):
            si = self.ui.InsertStretchSpacer(idx)
            self.sizerItems.insert(idx, (child, si))

    def removeChild(self, idx, child):
        self.relayout = True

        from .layout import Spacer
        if isinstance(child, WxBaseLayout):
            self.ui.Detach(idx)
            self.sizerItems.pop(idx)
        elif isinstance(child, WxBaseWidget):
            self.ui.Detach(idx)
            self.sizerItems.pop(idx)
        elif isinstance(child, Spacer):
            self.ui.Detach(idx)
            self.sizerItems.pop(idx)

    def postSync(self):
        if self.relayout and self.ui:
            self.relayout = False
            self.ui.Layout()
            # self.ui.Fit(getWindow(self.parent))
        super().postSync()



from .. import *
from .base import *

class TextField(TkBaseWidget):
    def __init__(self, model, edit_model=None):
        super().__init__()
        self.model = model
        self.edit_model = edit_model
        self.editing = False

    def update(self, prev):
        model_value = str(self.model.value)
        if prev and prev.ui:
            self.editing = prev.editing
            self.variable = prev.variable
            self.ui = prev.ui
            self.curr_value = prev.curr_value
            if self.curr_value.set(model_value) and not self.editing:
                self.variable.set(model_value)
        else:
            self.variable = tk.StringVar(self.tkparent.inner, model_value)
            self.variable.trace_add("write", self.on_variable_changed)
            self.curr_value = Prop(model_value)
            self.ui = tk.Entry(self.tkparent.inner, textvariable=self.variable)
            self.ui.bind('<Return>', self.on_return)
            self.ui.bind('<FocusOut>', self.on_focus_out)

        if self.edit_model and not self.editing:
            self.edit_model.value = model_value

        super().update(prev)

    def on_variable_changed(self, var, index, mode): # editing
        node = self.get_node()
        node.editing = True
        if node.edit_model:
            node.edit_model.value = self.variable.get()
        self._input()

    def on_change(self): # finish editing
        node = self.get_node()
        node.editing = False
        value = self.variable.get()
        node.model.value = value
        if node.edit_model:
            node.edit_model.value = value
        self._change()

    def on_return(self, e):
        self.on_change()

    def on_focus_out(self, e):
        self.on_change()



from .. import *
from .base import *

class Tabs(FBase):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    pui_terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ft.Tabs(tabs=[], expand=1)
        super().update(prev)

    def addChild(self, idx, child):
        tab = ft.Tab(text=child.parent.label, content=child.outer)
        self.ui.tabs.insert(idx, tab)

    def removeChild(self, idx, child):
        self.ui.tabs.pop(idx)

    def postSync(self):
        for i,c in enumerate(self.children):
            self.ui.tabs[i].text = c.label
        try:
            self.ui.update()
        except:
            pass
        return super().postSync()


class Tab(PUINode):
    pui_virtual = True
    def __init__(self, label):
        super().__init__()
        self.label = label



from .. import *
from .base import *
from PySide6.QtCore import Qt, QModelIndex, QAbstractItemModel

# XXX
# If click handler triggers a model reset, dblclick handler will not be called

class QAbstractItemModelAdapter(QtCore.QAbstractItemModel):
    def __init__(self, model: "BaseTreeAdapter"):
        super().__init__()
        self.model = model
        self.node = None

    def index(self, row, column, parent = QtCore.QModelIndex()):
        parent_node = parent.internalPointer() if parent.isValid() else None
        if 0 <= row and row < self.model.rowCount(parent_node):
            child = self.model.child(parent_node, row)
            return self.createIndex(row, column, child)
        return QtCore.QModelIndex()

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.ItemIsDropEnabled

        defaultFlags = super().flags(index)

        return defaultFlags | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

    def canDropMimeData(self, data, action, row, column, parent):
        if parent.isValid():
            pass
        else:
            return bool(self.node._onDropped)

    def dropMimeData(self, data, action, row, column, parent):
        if parent.isValid():
            pass
        else:
            event = QtGui.QDropEvent(QtCore.QPoint(0,0), action, data, QtCore.Qt.MouseButton.LeftButton, QtCore.Qt.KeyboardModifier.NoModifier, QtCore.QEvent.Drop)
            self.node._onDropped[0](event, *self.node._onDropped[1], **self.node._onDropped[2])

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()
        node = index.internalPointer()
        parent_node = self.model.parent(node)
        if parent_node:
            return self.createIndex(0, 0, parent_node)
        return QModelIndex()

    def data(self, index, role):
        if not index.isValid():
            return None
        node = index.internalPointer()
        if role == QtCore.Qt.DisplayRole:
            return self.model.data(node)
        return None

    def rowCount(self, parent):
        parent_node = parent.internalPointer() if parent.isValid() else None
        return self.model.rowCount(parent_node)

    def columnCount(self, parent):
        return 1

    def hasChildren(self, parent):
        parent_node = parent.internalPointer() if parent.isValid() else None
        return self.model.rowCount(parent_node) > 0

    def clicked(self, node):
        self.model.clicked(node)

    def dblclicked(self, node):
        self.model.dblclicked(node)

    def expanded(self, node):
        self.model.expanded(node)

    def collapsed(self, node):
        self.model.collapsed(node)

class QTreeNodeModelAdapter(QtCore.QAbstractItemModel):
    def __init__(self):
        super().__init__()
        self.node = None

    def index(self, row, column, parent = QtCore.QModelIndex()):
        parent_node = parent.internalPointer() if parent.isValid() else self.node
        if 0 <= row and row < len(parent_node.children):
            child = parent_node.children[row]
            return self.createIndex(row, column, child)
        return QtCore.QModelIndex()

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.ItemIsDropEnabled

        defaultFlags = super().flags(index)

        return defaultFlags | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

    def canDropMimeData(self, data, action, row, column, parent):
        if parent.isValid():
            pass
        else:
            return bool(self.node._onDropped)

    def dropMimeData(self, data, action, row, column, parent):
        if parent.isValid():
            pass
        else:
            event = QtGui.QDropEvent(QtCore.QPoint(0,0), action, data, QtCore.Qt.MouseButton.LeftButton, QtCore.Qt.KeyboardModifier.NoModifier, QtCore.QEvent.Drop)
            self.node._onDropped[0](event, *self.node._onDropped[1], **self.node._onDropped[2])

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()
        node = index.internalPointer()
        parent_node = node.parent
        if not isinstance(parent_node, TreeNode):
            parent_node = None
        if parent_node:
            return self.createIndex(0, 0, parent_node)
        return QModelIndex()

    def data(self, index, role):
        if not index.isValid():
            return None
        node = index.internalPointer()
        if role == QtCore.Qt.DisplayRole:
            return node.data
        return None

    def rowCount(self, parent):
        parent_node = parent.internalPointer() if parent.isValid() else self.node
        return len(parent_node.children)

    def columnCount(self, parent):
        return 1

    def hasChildren(self, parent):
        parent_node = parent.internalPointer() if parent.isValid() else self.node
        return len(parent_node.children) > 0

    def clicked(self, node):
        node._clicked(None)

    def dblclicked(self, node):
        node._dblclicked(None)

    def expanded(self, node):
        node._expanded()

    def collapsed(self, node):
        node._collapsed()

class Tree(QtBaseWidget):
    def __init__(self, model=None):
        super().__init__()
        self.layout_weight = 1
        self.model = model
        self.curr_model = None
        self.pendings = []
        self._expand_callback = None
        self._collapse_callback = None

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.qt_model = prev.qt_model
            self.curr_model = prev.curr_model

            self.ui.clicked.disconnect()
            self.ui.doubleClicked.disconnect()
            self.ui.expanded.disconnect()
            self.ui.collapsed.disconnect()
        else:
            self.qt_model = None
            self.curr_model = Prop()
            self.ui = QtWidgets.QTreeView()
            self.ui.setHeaderHidden(True)

        self.ui.clicked.connect(self.on_item_clicked)
        self.ui.doubleClicked.connect(self.on_item_double_clicked)
        self.ui.expanded.connect(self.on_item_expanded)
        self.ui.collapsed.connect(self.on_item_collapsed)

        if self.model:
            if self.curr_model.set(self.model):
                self.qt_model = QAbstractItemModelAdapter(self.model)
                self.qt_model.node = self
                self.ui.setModel(self.qt_model)
            else:
                self.qt_model.modelReset.emit()
        else:
            if not self.qt_model:
                self.qt_model = QTreeNodeModelAdapter()
                self.qt_model.node = self
                self.ui.setModel(self.qt_model)
            else:
                self.qt_model.beginResetModel()
                self.qt_model.node = self
                self.qt_model.endResetModel()

        for pending in self.pendings:
            pending[0](*pending[1:])
        self.pendings = []

        super().update(prev)

    def expandAll(self):
        if self.ui:
            self.ui.expandAll()
        else:
            self.pendings.append([self.expandAll])
        return self

    def collapseAll(self):
        if self.ui:
            self.ui.collapseAll()
        else:
            self.pendings.append([self.collapseAll])
        return self

    def expandable(self, enabled):
        if self.ui:
            self.ui.setItemsExpandable(enabled)
        else:
            self.pendings.append([self.expandable, enabled])
        return self

    def expand(self, cb, *args, **kwargs):
        self._expand_callback = (cb, args, kwargs)
        return self

    def _expanded(self):
        if self._expand_callback:
            cb, args, kwargs = self._expand_callback
            cb(*args, **kwargs)

    def collapse(self, cb, *args, **kwargs):
        self._collapse_callback = (cb, args, kwargs)
        return self

    def _collapsed(self):
        if self._collapse_callback:
            cb, args, kwargs = self._collapse_callback
            cb(*args, **kwargs)

    def on_item_clicked(self, index):
        treenode = index.internalPointer()
        self.get_node().qt_model.clicked(treenode)

    def on_item_double_clicked(self, index):
        treenode = index.internalPointer()
        self.get_node().qt_model.dblclicked(treenode)

    def on_item_expanded(self, index):
        treenode = index.internalPointer()
        self.get_node().qt_model.expanded(treenode)

    def on_item_collapsed(self, index):
        treenode = index.internalPointer()
        self.get_node().qt_model.collapsed(treenode)

class TreeNode(PUINode):
    def __init__(self, data=""):
        super().__init__()
        self._set_callback = None
        self.data = data
        self._expand_callback = None
        self._collapse_callback = None

    def expand(self, cb, *args, **kwargs):
        self._expand_callback = (cb, args, kwargs)
        return self

    def _expanded(self):
        if self._expand_callback:
            cb, args, kwargs = self._expand_callback
            cb(*args, **kwargs)

    def collapse(self, cb, *args, **kwargs):
        self._collapse_callback = (cb, args, kwargs)
        return self

    def _collapsed(self):
        if self._collapse_callback:
            cb, args, kwargs = self._collapse_callback
            cb(*args, **kwargs)



from .. import *
from .base import *

class ClickableQLabel(QtWidgets.QLabel):
    clicked  = QtCore.Signal()

    def mousePressEvent(self, ev):
        self.clicked.emit()
        super().mousePressEvent(ev)

class Label(QtBaseWidget):
    def __init__(self, text, selectable=False):
        super().__init__()
        self.text = str(text)
        self.selectable = selectable

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.setText(self.text)
        else:
            self.ui = ClickableQLabel(self.text)
            self.ui.setTextFormat(QtCore.Qt.TextFormat.PlainText)
            self.ui.clicked.connect(self._clicked)
        if self._onClicked:
            self.ui.setCursor(QtCore.Qt.PointingHandCursor)

        if self.selectable:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        else:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.NoTextInteraction)

        super().update(prev)



from .base import *

class MatplotlibCanvas(QtBaseWidget):
    def __init__(self, plotter, *args, **kwargs):
        super().__init__()
        self.plotter = plotter
        self.args = args
        self.kwargs = kwargs

    def update(self, prev):
        from matplotlib.backends.backend_qtagg import FigureCanvas
        from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
        from matplotlib.backends.qt_compat import QtWidgets
        from matplotlib.figure import Figure
        if prev and prev.ui:
            self.figure = prev.figure
            self.ui = prev.ui
        else:
            self.figure = Figure()
            self.ui = FigureCanvas(self.figure)
        self.plotter(self.figure, *self.args, **self.kwargs)
        self.ui.draw()
        super().update(prev)



from .. import *
from .base import *

class Tabs(TBase):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    pui_terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition
        self.tabs = None
        self.current_tab = None
        self._dirty = False

    def update(self, prev):
        if prev and prev.ui:
            self.current_tab = prev.current_tab
            self.ui = prev.ui
            self.tabhost = prev.tabhost
            self.frame = prev.frame
            self.tabs = prev.tabs
            self.mounted = prev.mounted
        else:
            self.ui = containers.Vertical()
            self.tabhost = containers.Container()
            self.frame = containers.Container()
            self.mounted = False
        super().update(prev)

    def addChild(self, idx, child):
        if not self.mounted:
            self.ui.mount(self.tabhost)
            self.ui.mount(self.frame)
            self.mounted = True
        self._dirty = True
        child.outer.display = False
        self.frame.mount(child.outer, before=idx)


    def removeChild(self, idx, child):
        self.ui.removeTab(idx)
        self._dirty = True
        child.tremove()

    def postSync(self):
        if self._dirty:
            if self.tabs:
                self.tabs.remove()
            tabs = []
            for c in self.children:
                tab = widgets.Tab(c.label)
                tab.puinode = c
                tabs.append(tab)
            self.tabs = widgets.Tabs(*tabs)
            self.tabs.puinode = self
            self.tabhost.mount(self.tabs)
            self._dirty = False
        super().postSync()

    def _tab_activated(self, event: widgets.Tabs.TabActivated):
        node = self.get_node()
        if node.current_tab:
            node.current_tab.outer.display = False
        if event.tab:
            tab = event.tab.puinode.get_node()
            tab.outer.display = True
            self.current_tab = tab

class Tab(PUINode):
    pui_virtual = True
    def __init__(self, label):
        super().__init__()
        self.label = label



from .. import *
from .base import *

from PySide6 import QtWidgets, QtGui
from PySide6.QtGui import QPainter, QColor, QPainterPath, QImage

class PUIQtCanvas(QtWidgets.QWidget):
    def __init__(self, node, width=None, height=None):
        self.puinode = node
        self.width = width
        self.height = height
        super().__init__()

    def minimumSizeHint(self):
        return QtCore.QSize(self.width, self.height)

    def mouseDoubleClickEvent(self, event):
        e = PUIEvent()
        e.button = event.button().value
        e.x, e.y = event.position().toPoint().toTuple()
        self.puinode._dblclicked(e)

    def mousePressEvent(self, event):
        e = PUIEvent()
        e.button = event.button().value
        e.x, e.y = event.position().toPoint().toTuple()
        self.puinode._mousedown(e)

    def mouseReleaseEvent(self, event):
        e = PUIEvent()
        e.button = event.button().value
        e.x, e.y = event.position().toPoint().toTuple()
        self.puinode._mouseup(e)

    def mouseMoveEvent(self, event):
        e = PUIEvent()
        e.button = event.button().value
        e.x, e.y = event.position().toPoint().toTuple()
        self.puinode._mousemove(e)

    def wheelEvent(self, event):
        e = PUIEvent()
        e.x, e.y = event.position().toPoint().toTuple()
        e.y_delta = event.pixelDelta().y()
        e.x_delta = event.pixelDelta().x()
        e.v_delta = event.angleDelta().y()
        e.h_delta = event.angleDelta().x()
        modifier = 0
        emodifiers = event.modifiers()
        if emodifiers & QtCore.Qt.ShiftModifier:
            modifier |= KeyModifier.SHIFT
        if emodifiers & QtCore.Qt.ControlModifier:
            modifier |= KeyModifier.CTRL
        if emodifiers & QtCore.Qt.AltModifier:
            modifier |= KeyModifier.ALT
        if emodifiers & QtCore.Qt.MetaModifier:
            modifier |= KeyModifier.META
        e.modifiers = modifier
        self.puinode._wheel(e)

    def paintEvent(self, event):
        puinode = self.puinode.get_node()
        puinode.qpainter = QPainter()
        puinode.qpainter.begin(self)
        puinode.qpainter.setRenderHints(QtGui.QPainter.Antialiasing, True)

        if not puinode.style_bgcolor is None:
            bgBrush = QtGui.QBrush()
            bgBrush.setColor(QtGui.QColor(puinode.style_bgcolor))
            bgBrush.setStyle(QtCore.Qt.SolidPattern)
            rect = QtCore.QRect(0, 0, self.width or self.geometry().width(), self.height or self.geometry().height())
            puinode.qpainter.fillRect(rect, bgBrush)

        puinode.width = self.geometry().width()
        puinode.height = self.geometry().height()
        immediate = puinode.painter(puinode, *puinode.args)
        puinode.qpainter.end()
        puinode.qpainter = None
        if immediate:
            self.update()

class ImageResource():
    @staticmethod
    def load(path):
        ir = ImageResource()
        ir.qimage = QImage(path)
        return ir

    def crop(self, x, y, width, height):
        ir = ImageResource()
        ir.qimage = self.qimage.copy(x, y, width, height)
        return ir

    def scale(self, width, height, keepAspectRatio=True, quality=0):
        ir = ImageResource()
        method = {
            0: QtCore.Qt.TransformationMode.FastTransformation
        }.get(quality, QtCore.Qt.TransformationMode.SmoothTransformation)
        ir.qimage = self.qimage.scaled(width, height, QtCore.Qt.AspectRatioMode.KeepAspectRatio if keepAspectRatio else QtCore.Qt.IgnoreAspectRatio, method)
        return ir

class Canvas(QtBaseWidget):
    def __init__(self, painter, *args):
        super().__init__()
        self.ui = None
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.puinode = self
            self.ui.width = self.layout_width or 0
            self.ui.height = self.layout_height or 0
        else:
            self.ui = PUIQtCanvas(self, self.layout_width or 0, self.layout_height or 0)
        self.ui.setMouseTracking(bool(self._onMouseMove))
        self.ui.update()
        super().update(prev)

    def drawText(self, x, y, text, w=None, h=None, size=12, color=None, rotate=0, anchor=Anchor.LEFT_TOP):
        if w is None:
            w = self.ui.geometry().width()
        if h is None:
            h = self.ui.geometry().height()
        self.qpainter.save()
        br = self.qpainter.boundingRect(0, 0, w, h, 0, text)
        self.qpainter.restore()

        dx = 0
        dy = 0
        if anchor.value[0]=="center":
            dx = br.width()/2
        elif anchor.value[0]=="right":
            dx = br.width()

        if anchor.value[1]=="center":
            dy = br.height()/2
        elif anchor.value[1]=="bottom":
            dy = br.height()

        self.qpainter.save()
        self.qpainter.translate(int(x+dx), int(y+dy))
        self.qpainter.rotate(rotate)
        self.qpainter.translate(int(-dx), int(-dy))

        if size is not None:
            font = self.qpainter.font()
            font.setPointSize(size)
            self.qpainter.setFont(font)

        if color is not None:
            self.qpainter.setPen(QColor(color))

        self.qpainter.drawText(0, 0, w, h, 0, text)
        self.qpainter.restore()

    def drawLine(self, x1, y1, x2, y2, color=None, width=1):
        self.qpainter.save()

        pen = self.qpainter.pen()
        if color is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(color))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawLine(x1, y1, x2, y2)

        self.qpainter.restore()

    def drawPolyline(self, coords, color=None, width=1):
        self.qpainter.save()

        pen = self.qpainter.pen()
        if color is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(color))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawPolyline([QtCore.QPointF(x,y) for x,y in coords])

        self.qpainter.restore()

    def drawPolygon(self, coords, fill=None, stroke=None, width=1):
        self.qpainter.save()

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        polygon = QtGui.QPolygonF()
        for p in coords:
            polygon.append(QtCore.QPointF(*p))
        self.qpainter.drawPolygon(polygon)

        self.qpainter.restore()

    def drawRect(self, x1, y1, x2, y2, fill=None, stroke=None, width=1):
        self.qpainter.save()

        x = min(x1, x2)
        y = min(y1, y2)
        w = abs(x2-x1)
        h = abs(y2-y1)

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawRect(x, y, w, h)

        self.qpainter.restore()

    def drawEllipse(self, x, y, rx, ry, fill=None, stroke=None, width=1):
        self.qpainter.save()

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self.qpainter.drawEllipse(QtCore.QPointF(x,y), rx, ry)

        self.qpainter.restore()

    def drawShapely(self, shape, fill=None, stroke=None, width=1):
        self.qpainter.save()

        brush = self.qpainter.brush()
        if fill is None:
            brush.setStyle(QtCore.Qt.NoBrush)
        else:
            brush.setStyle(QtCore.Qt.SolidPattern)
            brush.setColor(QColor(fill))
        self.qpainter.setBrush(brush)

        pen = self.qpainter.pen()
        if stroke is None:
            pen.setStyle(QtCore.Qt.NoPen)
        else:
            pen.setStyle(QtCore.Qt.SolidLine)
            pen.setColor(QColor(stroke))
            pen.setWidth(width)
        self.qpainter.setPen(pen)

        self._drawShapely(shape, fill, stroke, width)

        self.qpainter.restore()


    def _drawShapely(self, shape, fill=None, stroke=None, width=1):
        if hasattr(shape, "geoms"):
            for g in shape.geoms:
                self.drawShapely(g, fill, stroke, width)
        elif hasattr(shape, "exterior"): # polygon
            path = QPainterPath()

            exterior = QtGui.QPolygonF()
            for p in shape.exterior.coords:
                exterior.append(QtCore.QPointF(*p))
            path.addPolygon(exterior)

            for h in shape.interiors:
                hole = QtGui.QPolygonF()
                for p in h.coords:
                    hole.append(QtCore.QPointF(*p))
                hpoly = QPainterPath()
                hpoly.addPolygon(hole)
                path = path.subtracted(hpoly)

            self.qpainter.drawPath(path)
        elif hasattr(shape, "x") and hasattr(shape, "y"): # point
            self.drawEllipse(shape.x, shape.y, width/2, width/2, fill=stroke)
        elif hasattr(shape, "coords"): # linestring, linearring
            self.drawPolyline(shape.coords, color=stroke, width=width)
        else:
            raise RuntimeError(f"Not implemented: drawShapely({type(shape).__name__}) {dir(shape)}")

    def loadImage(self, image_path):
        return ImageResource.load(image_path)

    def drawImage(self, image, x=0, y=0, width=None, height=None, src_x=0, src_y=0, src_width=None, src_height=None, opacity=1.0):
        if image.qimage.isNull():
            return

        if src_width is None:
            src_width = image.qimage.width() - src_x
        if src_height is None:
            src_height = image.qimage.height() - src_y

        source_rect = QtCore.QRect(src_x, src_y, src_width, src_height)

        if width is None:
            width = src_width
        if height is None:
            height = src_height

        dest_rect = QtCore.QRect(x, y, width, height)
        self.qpainter.setOpacity(opacity)
        self.qpainter.drawImage(dest_rect, image.qimage, source_rect)
        self.qpainter.setOpacity(1.0)


from .. import *
from .base import *

class VBox(TBase):
    container_y = True
    expand_x_prio = 1
    expand_y_prio = 2
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.Vertical()
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer, before=idx)

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()

class HBox(TBase):
    container_x = True
    expand_x_prio = 2
    expand_y_prio = 1
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.Horizontal()
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer, before=idx)

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()

class Spacer(TBase):
    def __init__(self):
        super().__init__()
        self.layout_weight = 1

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = widgets.Static("")
        super().update(prev)


from .. import *
from .base import *

class QtTableModelAdapter(QtCore.QAbstractTableModel):
    def __init__(self, model: "BaseTableAdapter"):
        super().__init__()
        self.model = model

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            return self.model.data(index.row(), index.column())
        elif role == QtCore.Qt.EditRole:
            return self.model.editData(index.row(), index.column())

    def setData(self, index, value, role):
        if role == QtCore.Qt.EditRole:
            self.model.setData(index.row(), index.column(), value)
            return True

    def flags(self, index):
        flags = super().flags(index)
        if self.model.editable(index.row(), index.column()):
            flags |= QtCore.Qt.ItemIsEditable
        return flags

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                if hasattr(self.model, "columnHeader"):
                    if self.model.columnHeader is None:
                        return None
                    else:
                        return self.model.columnHeader(section)
                else:
                    return super().headerData(section, orientation, role)
            else:
                if hasattr(self.model, "rowHeader"):
                    if self.model.rowHeader is None:
                        return None
                    else:
                        return self.model.rowHeader(section)
                else:
                    return super().headerData(section, orientation, role)

    def rowCount(self, index):
        return self.model.rowCount()

    def columnCount(self, index):
        return self.model.columnCount()


class QtTableNodeModelAdapter(QtCore.QAbstractTableModel):
    def __init__(self, children):
        super().__init__()
        self.children = children

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            data = self.children[index.row()].children[index.column()].data
            if isinstance(data, BaseBinding):
                return str(data.value)
            return data
        return None

    def setData(self, index, value, role):
        if role == QtCore.Qt.EditRole:
            n = self.children[index.row()].children[index.column()]
            if isinstance(n.data, BaseBinding):
                n.data.value = value
            else:
                n._set(value)
            return True

    def flags(self, index):
        flags = super().flags(index)
        n = self.children[index.row()].children[index.column()]
        if isinstance(n.data, BaseBinding) or n._set_callback:
            flags |= QtCore.Qt.ItemIsEditable
        return flags

    def rowCount(self, index):
        return len(self.children)

    def columnCount(self, index):
        if self.children:
            return len(self.children[0].children)
        return 0

class Table(QtBaseWidget):
    def __init__(self, model=None, autofit=True):
        super().__init__()
        self.layout_weight = 1
        self.model = model
        self.autofit = autofit
        self.curr_model = None

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.qt_model = prev.qt_model
            self.curr_model = prev.curr_model
        else:
            self.qt_model = None
            self.curr_model = Prop()
            self.ui = QtWidgets.QTableView()

        if self.model:
            if self.model.columnHeader is None:
                self.ui.horizontalHeader().hide()
            else:
                self.ui.horizontalHeader().show()

            if self.model.rowHeader is None:
                self.ui.verticalHeader().hide()
            else:
                self.ui.verticalHeader().show()

            if self.curr_model.set(self.model):
                self.qt_model = QtTableModelAdapter(self.model)
                self.ui.setModel(self.qt_model)
            else:
                self.qt_model.refresh()
        else:
            self.ui.horizontalHeader().hide()
            self.ui.verticalHeader().hide()
            self.qt_model = QtTableNodeModelAdapter(self.children)
            self.ui.setModel(self.qt_model)

        if self.autofit:
            self.ui.resizeColumnsToContents()

        super().update(prev)

class TableNode(PUINode):
    def __init__(self, data=""):
        super().__init__()
        self._set_callback = None
        self.data = data

    def set(self, cb, *args, **kwargs):
        self._set_callback = (cb, args, kwargs)
        return self

    def _set(self, data):
        if self._set_callback:
            cb, args, kwargs = self._set_callback
            cb(data, *args, **kwargs)
